
--[[
=head1 NAME

applets.Tetris.TetrisApplet - Tetris Game

=head1 DESCRIPTION

Tetris is a simple game for the Jive platform

=head1 FUNCTIONS

Applet related methods are described in L<jive.Applet>. TetrisApplet overrides the
following methods:

=cut
--]]


-- stuff we use
local ipairs, tostring = ipairs, tostring
local math, table = math, table

local oo               = require("loop.simple")

local Applet           = require("jive.Applet")
local Window           = require("jive.ui.Window")
local Surface          = require("jive.ui.Surface")
local Popup            = require("jive.ui.Popup")
local Icon             = require("jive.ui.Icon")
local Framework        = require("jive.ui.Framework")
local Audio	       = require("jive.ui.Audio")

local log              = require("jive.utils.log").addCategory("Tetris", jive.utils.log.DEBUG)

local color_index 	= 1
local current_index     = 2
local LINES_PER_LEVEL   = 5
local lines_to_next     = LINES_PER_LEVEL
local current_level     = 1
local tick_counter	= 0
local drop_flag         = false

local current_block     = {}
local current_state 	= 1

local EVENT_SCROLL     = jive.ui.EVENT_SCROLL
local EVENT_KEY_PRESS  = jive.ui.EVENT_KEY_PRESS
local EVENT_KEY_DOWN   = jive.ui.EVENT_KEY_DOWN
local KEY_BACK         = jive.ui.KEY_BACK
local KEY_LEFT         = jive.ui.KEY_LEFT
local KEY_RIGHT        = jive.ui.KEY_RIGHT
local KEY_UP           = jive.ui.KEY_UP
local KEY_DOWN         = jive.ui.KEY_DOWN
local KEY_ADD	       = jive.ui.KEY_ADD
local KEY_PLAY	       = jive.ui.KEY_PLAY
local KEY_GO	       = jive.ui.KEY_GO

local EVENT_MOUSE_DOWN       = jive.ui.EVENT_MOUSE_DOWN
local EVENT_MOUSE_PRESS      = jive.ui.EVENT_MOUSE_PRESS

local FRAMERATE        = jive.ui.FRAME_RATE
local BLOCKSIZE        = 16
local srf = nil
local window = nil
local w, h = Framework:getScreenSize()

local temp_field       = {}
local playField	       = {}
local score = 0

local transformation_data = {
			
	--Data for rotating block between states. New Coords for each of 4 blocks, followed by value of the next state.
	--Should be usable for rotating blocks CCW.

{
	--Color 1
	{ { 1 ,  0} , { 1 ,  1 } , { 1 ,  2}, { 0 ,  2}, 2 }, -- state 1		
	{ { 1 ,  1} , { 0 ,  1 } , {-1 ,  1}, {-1 ,  0}, 3 }, -- state 2		
	{ {-1 ,  1} , {-1 ,  0 } , {-1 , -1}, { 0 , -1}, 4 }, -- state 3	
	{ {-1 , -2} , { 0 , -2 } , { 1 , -2}, { 1 , -1}, 1 }  --state 4
	},
	--Color 2
	{
	{ { 1 ,  0} , { 1 ,  1 }, { 1 ,  2}, { 0 ,  0}, 2 }, -- state 1
	{ { 1 ,  1} , { 0 ,  1 }, {-1 ,  1}, { 1 ,  0}, 3 }, -- state 2
	{ {-1 ,  1} , {-1 ,  0 }, {-1 , -1}, { 0 ,  1}, 4 }, -- state 3
	{ {-1 , -2} , { 0 , -2 }, { 1 , -2}, {-1 , -1}, 1 }, -- state 4
	},	
	--Color 3
	{
	{ { 1 ,  0} , { 1 ,  1 }, { 1 ,  2}, {0 ,  1}, 2 }, -- state 1
	{ { 1 ,  1} , { 0 ,  1 }, {-1 ,  1}, {0 ,  0}, 3 }, -- state 2
	{ {-1 ,  1} , {-1 ,  0 }, {-1 , -1}, {0 ,  0}, 4 }, -- state 3
	{ {-1 , -2} , { 0 , -2 }, { 1 , -2}, {0 , -1}, 1 } -- state 4
	},
	--Color 4
	{
	{ { 1 , 0} , {1 , 1}, {0 , 1}, {0 , 2}, 2 }, -- state 1
	{ {-1 , 0} , {0 , 0}, {0 , 1}, {1 , 1}, 1 }  --state 2
	},
	--Color 5
	{},
	--Color 6
	{
	{ {0 ,  1} , {0 , 2}, {-1 , 1}, {-1 , 0}, 2 }, -- state 1
	{ {0 , -1} , {1, -1}, { 0 , 0}, {-1 , 0}, 1 }  -- state 2
	},
	--Color 7
	{
	{ {0 ,  1} , {0 , 2}, {0 ,  3}, { 0 ,  0}, 2 }, -- state 1 
	{ {0 , -1} , {1, -1}, {2 , -1}, {-1 , -1}, 1 }  -- state 2
	}
}

-- Color data.
local cyan   = 0x00F0F0F0
local red    = 0xFF0000FF
local purple = 0xFF00FFFF
local green  = 0x00FF00FF
local white  = 0xFFFFFFFF
local yellow = 0xFFFF00FF
local orange = 0xFF9900FF
local color_wheel = { cyan, red, purple, green, white, yellow, orange  }

local minX = 1
local minY = 1
local maxX = math.floor(w / BLOCKSIZE) - 2
local maxY = math.floor((h-30) / BLOCKSIZE) - 2
local labelStatus = nil
local spawn_location_x =  math.floor(maxX / 2)
local spawn_location_y = 0;
local relative_spawn_offsets =
	-- Spawn locations for each of the 7 types of blocks. Relative to spawn_location_x and _y.
	{ 	{ { 0 , 1 } , { 1 , 1 } , { 2 , 1 } , { 2 , 2 } }, -- color 1 
		{ { 0 , 1 } , { 1 , 1 } , { 2 , 1 } , { 0 , 2 } }, -- color 2
		{ { 0 , 1 } , { 1 , 1 } , { 2 , 1 } , { 1 , 2 } }, -- color 3
		{ { 0 , 1 } , { 1 , 1 } , { 1 , 2 } , { 2 , 2 } }, -- color 4
		{ { 0 , 1 } , { 1 , 1 } , { 0 , 2 } , { 1 , 2 } }, -- color 5
		{ { 0 , 1 } , { 1 , 1 } , { 0 , 2 } , {-1 , 2 } }, -- color 6
		{ { 0 , 1 } , { 1 , 1 } , { 2 , 1 } , {-1 , 1 } }  -- color 7
	}

module(...)
oo.class(_M, Applet)

function displayName(self)
	return "Tetris"
end

	--randomly spawn a block (one of the 7 types)

function spawnRandom(self)
	
	local random_color = math.random( 1 , 7 )	
	current_state = 1
	local is_clear = true
	current_block = {}

	for i = 1 , #relative_spawn_offsets[random_color] do
		if playField[spawn_location_x + relative_spawn_offsets[random_color][i][1]][spawn_location_y + relative_spawn_offsets[random_color][i][2]][color_index] != 0 then
			is_clear = false
			gameOver(self)
		end
	end
	
	if is_clear then

		for i = 1,  #relative_spawn_offsets[random_color] do
			playField[spawn_location_x + relative_spawn_offsets[random_color][i][1]][spawn_location_y + relative_spawn_offsets[random_color][i][2]] = { random_color , 1 }
		table.insert ( current_block , { spawn_location_x + relative_spawn_offsets[random_color][i][1] , spawn_location_y + relative_spawn_offsets[random_color][i][2]} )
		end
	end
end

	--transfer the current block from temp_field to playField. nil out values in temp_field afterwards.
function currentFromBuffer(self)
	for i = 1,#current_block do 		
		playField[current_block[i][1]][current_block[i][2]] = temp_field[current_block[i][1]][current_block[i][2]]
		temp_field[current_block[i][1]][current_block[i][2]] = nil
	end
end
	--erase the current block from playField
function clearCurrent(self)
	for i = 1,#current_block do
		playField[current_block[i][1]][current_block[i][2]] = {0, 0}
	end
end

	--rotate the block based on the current type of block and state, relative to the current position. Uses transformation_data.

function rotateByRelativeMatrix(rotate_color, rotate_state, rotate_direction)

	local valid_transform = true
	local first_x = current_block[1][1]
	local first_y = current_block[1][2]
	for i = 1, #transformation_data[rotate_color][rotate_state] - 1 do
		local newX = first_x + transformation_data[rotate_color][rotate_state][i][1]
		local newY = first_y + transformation_data[rotate_color][rotate_state][i][2]
		if (newX > maxX or newX < 1 or newY > maxY or newY < 1) or (playField[newX][newY][color_index]!= 0 and playField[newX][newY][current_index] == 0 ) then
			valid_transform = false
			break
		end
	end
	if valid_transform == true then
		for i = 1, # transformation_data[rotate_color][rotate_state] - 1 do
			temp_field[first_x + transformation_data[rotate_color][rotate_state][i][1]][first_y + transformation_data[rotate_color][rotate_state][i][2]] = playField[current_block[i][1]][current_block[i][2]]
		end
		clearCurrent(self)
		current_block = {}
		for i = 1, #transformation_data[rotate_color][rotate_state] - 1 do
			table.insert( current_block, {first_x + transformation_data[rotate_color][rotate_state][i][1] , first_y + transformation_data[rotate_color][rotate_state][i][2]} )
		end
		current_state = transformation_data[rotate_color][rotate_state][5]
		currentFromBuffer(self)
	end
end


function rotate(self)

	local color = playField[current_block[1][1]][current_block[1][2]][color_index]
	local first_x = current_block[1][1]
	local first_y = current_block[1][2]
	
	--don't bother rotating the 2x2 square.

	if color != 5 then 
		
		rotateByRelativeMatrix(color, current_state, 1)
	end
end

--initialize the playing field as blank. Reset score, level, lines.

function initPlayfield(self)
	
	lines_to_next   = LINES_PER_LEVEL
	current_level   = 1
	tick_counter	= 0
	score		= 0
	for i = 1, maxX do
		playField[i] = {}
		temp_field[i] = {}
		for j = 1, maxY do
			playField[i][j] = { 0 , 0 }
		end
	end
	spawnRandom()

end

function openWindow(self)
	self.window = self:_window("Tetris")
	self:tieAndShowWindow(self.window)
	return self.window
end

	-- draw a block of blocksize from x, y with color.

function drawBlock(self, x , y , color)

  	srf:filledRectangle(BLOCKSIZE*x,BLOCKSIZE*y, BLOCKSIZE*(x +1)-1, BLOCKSIZE*(y+1)-1, color)
end

function drawField(self)
	
	local lStatus
	if gameover == true then
		srf:filledRectangle(BLOCKSIZE * 3, BLOCKSIZE * 3, (maxX - 1) * BLOCKSIZE, (maxY - 1)* BLOCKSIZE, 0xFFFFFFFF)
		lStatus = Surface:drawText(window:styleFont("squeezebox"), 0x000000FF,  "You have lost. Your score: " .. score)
	else

		lStatus = Surface:drawText(window:styleFont("squeezebox"), 0x000000FF,  "Score: " .. score .. " Level: " .. current_level)
		--draw the border
		srf:filledRectangle(0, 0 , w, h, 0xFFFFFFFFF)
		srf:filledRectangle(BLOCKSIZE, BLOCKSIZE, (maxX + 1) * BLOCKSIZE, (maxY + 1 )* BLOCKSIZE, 0x000000FF)

		--draw the field
		for i = 1, maxX do
			for j = 1, maxY do
				if playField[i][j][color_index] != 0 then
					drawBlock(self, i , j , color_wheel[playField[i][j][color_index]])
				end
			end
		end
	end
	if labelStatus != nil then
		labelStatus:setValue(lStatus)
	else 
		labelStatus = Icon("lives", lStatus)
		window:addWidget(labelStatus)
	end
	self.bg:reDraw()
end

function skin(self, s)
	s.snake.layout = Window.noLayout
end


--game is over.

function gameOver(self)
	gameover = true
	drawField(self)
end

-- try to drop the current brick. If it hits something, stop it and check to see if any lines are cleared. Then spawn a new brick.

function dropCurrent(self)

	is_stuck = false

	for i = 1,#current_block do 
	
		if current_block[i][2] >= maxY then
			is_stuck = true

		elseif playField[current_block[i][1]][current_block[i][2]+1][color_index] != 0 and playField[current_block[i][1]][current_block[i][2]+1][current_index] != 1 then
			is_stuck = true
			break
		end

	end

	if is_stuck then
		drop_flag = false
		for i = 1,#current_block do 
			playField[current_block[i][1]][current_block[i][2]][current_index] = 0
		end
		dropLevel(self)
	else
		moveCurrent(0, 1)
	end

end

-- shift everything above "line" down one.

function clearLines(self, line)

	for j = 1, line - 1 do
		for i = 1, maxX do
			temp_field[i][j + 1] = playField[i][j]
		end
	end
	for j = 2, line do
		for i = 1, maxX do
			playField[i][j] = temp_field[i][j]
			temp_field[i][j] = nil
		end
	end
end

--check if any rows are complete. Clear them and give points.

function dropLevel(self)
	local cleared = 0	
	local row_complete = true
		for j = 1, maxY do
			for i = 1, maxX do
			if playField[i][j][color_index] == 0 then
				row_complete = false
				break
			end
		end
		if row_complete then
			clearLines(self, j)
			cleared = cleared + 1
			lines_to_next = lines_to_next - 1
			if (lines_to_next == 0) then
				lines_to_next = LINES_PER_LEVEL
				current_level = current_level + 1
			end
		else
			row_complete = true
		end
	end

	score = score + (cleared * cleared * 5 * current_level)

	spawnRandom(self)
	
end

-- move the current block by the x and y offsets (positive is right/down, negative is left/up)

function moveCurrent(x_offset, y_offset)

	if current_block then
		local is_clear = true
		for i = 1,#current_block do 
			if current_block[i][2] + y_offset > maxY or current_block[i][1] + x_offset < 1 or current_block[i][1] + x_offset > maxX or playField[current_block[i][1] + x_offset][current_block[i][2]+ y_offset][color_index] != 0 and playField[current_block[i][1] + x_offset][current_block[i][2] + y_offset][current_index] != 1 then
			is_clear = false
			end
		end
		if is_clear then
			for i = 1,#current_block do 		
				if temp_field[current_block[i][1] + x_offset] then
				else
					temp_field[current_block[i][1] + x_offset] = {}
				end
					temp_field[current_block[i][1] + x_offset][current_block[i][2] + y_offset] = playField[current_block[i][1]][current_block[i][2]]
					playField[current_block[i][1]][current_block[i][2]] = {0, 0}
					current_block[i][1] = current_block[i][1] + x_offset
					current_block[i][2] = current_block[i][2] + y_offset
				end
				for i = 1,#current_block do 		
					playField[current_block[i][1]][current_block[i][2]] = temp_field[current_block[i][1]][current_block[i][2]]
					temp_field[current_block[i][1]][current_block[i][2]] = nil
				end
			end
		end

end

--handle keys. If game is over, and a key is pressed then restart the game.

function handleKey(self)

		if gameover then
			gameover = false
			initPlayfield(self)
		end

		if direction == KEY_LEFT then
			moveCurrent(-1, 0)
		elseif direction == KEY_RIGHT then
			moveCurrent(1, 0)
		elseif direction == KEY_DOWN then
			drop_flag = true
		elseif direction == KEY_UP then
			rotate(self)
		end

end

-- each FRAMERATE - current_level ticks try to drop the current brick.

function autoMove(self) 
	
	tick_counter = tick_counter + 1
	if (tick_counter >= FRAMERATE - current_level) or drop_flag then

		dropCurrent(self)
		tick_counter = 0
	end

end

function keyEvent(self, evt)

	  local key = evt:getKeycode()
	  if key == KEY_LEFT then
			direction = KEY_LEFT
		elseif key == KEY_RIGHT then
			direction = KEY_RIGHT
		elseif key == KEY_DOWN then
			direction = KEY_DOWN
		elseif key == KEY_UP then
			direction = KEY_UP
		elseif key == KEY_ADD then
			direction = KEY_LEFT
		elseif key == KEY_GO then
			direction = KEY_DOWN
		elseif key == KEY_PLAY then
			direction = KEY_RIGHT
		end

	handleKey(self)
end

function scrollEvent(self, evt)

		local scroll = evt:getScroll()
		-- May change this eventually, to do CW and CCW rotation
		if scroll == 1 then
			 -- turn right
		 	direction = KEY_UP

		elseif scroll == -1 then
			-- turn left
		 	direction = KEY_UP
		end

		handleKey(self)
end


function _window(self, ...)	
  	window = Window(self:displayName())

	srf = Surface:newRGBA(w, h)
	srf:filledRectangle(0, 0, w, h, 0x000000FF)
	self.bg = Icon("background", srf)

	window:addListener(EVENT_KEY_PRESS, 
		function(evt)
			if evt:getKeycode() == KEY_BACK then
				window:hide()
			end
		end
	)

	window:addListener(EVENT_KEY_DOWN, 
		function(evt)
			keyEvent(self, evt)
		end
	)

	window:addListener(EVENT_SCROLL,
		function(evt)
			log:info("Scrollacious")
			scrollEvent(self, evt)
		end
	)

	window:addWidget(self.bg)

	initPlayfield()

	self.bg:addAnimation(
		function()
		  self:autoMove()
		  self:drawField()
		end,
		FRAMERATE	
	)

	return window
end


--[[

=head1 LICENSE

Copyright 2007 Logitech Harmony,

Based on "snake" by Lukas Frey

Feel free to reuse!

=cut
--]]
